/* $Workfile: BertExDemo1.cpp $   $Revision: 3 $    $Modtime: 1/10/06 4:44p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    BertExDemo.cpp

    A simple example console application demonstrating the use of the
    software BER Testing using OTX board's Burst device.

      Thor-8-PCI-Plus
      Sleipnir

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 1 Line - #0 which can be looped with an external loop cable or connected
    to remote BERT.

    Limitations:
    - Callback functions can be used only on 32-bit platforms;
      On 64-bit platforms normal work can be only in first 4Gb block of memory;

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "OtxData.h"

#include "OtxBertEx.h"

/*******************************************************************************
 Defines
*******************************************************************************/

#define DEMO_DEFAULT_LOG_FILE   "BertExDemo.log"
#define DEMO_DEFAULT_BOARD_NO   0

// Use these defines as the limit buffer size
#define DEMO_BURST_TX_BUF_SIZE  10000* 128   //0x8000    // 32k
#define DEMO_RX_BUF_SIZE        10000* 128   //0x8000

// The number of samples per one timeslot in DMA buffer
// Not: It should be changed manually to get optimal
// delay and CPU usage.
#define DEMO_RX_SAMPLES         200
#define DEMO_TX_SAMPLES         200

#define DEMO_MAX_DEV 256

/*******************************************************************************
 Structures
*******************************************************************************/
typedef struct DemoEnvT {
  // Common 
  char* pszInputStr;
  OTX_BOOL bExitOnError;
  char cwd[OTX_MAX_PATH];
  char logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int coLi;
  int coHighways;
  int nCurrentLi;
  int coBertEx;
  OtxT1E1LiModeE eT1E1Mode;    // T1 or E1 mode
  int boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS* pE1CfgData, OUT OtxT1ConfigOptionsS* pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS* pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char* formatStr, ...);
static void DemoLogf2(char* formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP);
static void DemoCreateLogicalDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
int DemoPrintBertEvent(OtxEventDataS* pEventData);
void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_UINT32 g_coTimeSlotMask = 32;

OTX_EVENT g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE g_hEventQueue = 0;
FILE* g_pDemoLogFile = NULL;
FILE* g_pBurstSendFile = NULL;
FILE* g_pBurstRecvFile = NULL;

DemoEnvT g_demoEnv;
OTX_BOOL g_bStatsThreadActive = OTX_FALSE;

OsaMutexS g_demoLogfCritSec;

struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hQuadLi[2];
  OTX_HANDLE hBurst;
} g_hPhysDevs;

struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;

struct {
  BERTX_HANDLE hHandle;
  OTX_INT32 nLi;
  OTX_UINT32 nTs;
  OTX_UINT32 nBurst;
} g_sqDspBertSender[DEMO_MAX_DEV];

struct {
  BERTX_HANDLE hHandle;
  OTX_INT32 nLi;
  OTX_UINT32 nTs;
  OTX_UINT32 nBurst;
} g_sqDspBertReceiver[DEMO_MAX_DEV];

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32 nSourceId;
  OTX_BOOL bVerbose;
  OTX_BOOL bActive;
  OTX_UINT8 sqTxBuf[DEMO_BURST_TX_BUF_SIZE];
  FILE* pTxFile;
  OTX_INT32 nTsToSend;                     // Default is -1 (All TS)
  OTX_UINT32 cbTxBuf;
  OTX_BOOL bFileLoop;
  BERTX_HANDLE hDev[DEMO_MAX_DEV];            // related BertEx devices
  OTX_UINT32 nDevs;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32 nSourceId;
  OTX_BOOL bVerbose;
  OTX_BOOL bActive;
  FILE* pRxFile;
  OTX_INT32 nTsToRecord;                   // Default is -1 (All TS)
  BERTX_HANDLE hDev[DEMO_MAX_DEV];
  OTX_UINT32 nDevs;
};

struct {
  struct sBurstSender mBurstSender[8];        
  struct sBurstReceiver mBurstReceiver[8];    
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;

typedef struct {
  char *sDescription;
  BERTPATTERN DspBertPattern;
} BERTPATTERNCTRL;

// Supported BERT pattern
BERTPATTERNCTRL arrPattern[] = {

  // PRBS patterns
  { "2^6-1",    { PATTERN_PRBS, 0x0030, 0x0000 } }, // 2^6-1    (taps [5,6])
  { "2^9-1",    { PATTERN_PRBS, 0x0110, 0x0000 } }, // 2^9-1 (taps [5,9])
  { "2^10-1",   { PATTERN_PRBS, 0x0220, 0x0000 } }, // 2^10-1 (taps [7,10])
//  { "2^9-1",    { PATTERN_PRBS, 0x0130, 0x0000 } }, // 2^9-1    (taps [5,9])
//  { "2^10-1",   { PATTERN_PRBS, 0x0200, 0x0000 } }, // 2^10-1   (taps [7,10])
  { "2^11-1",   { PATTERN_PRBS, 0x0500, 0x0000 } }, // 2^11-1   (taps [9,11])
  { "2^15-1",   { PATTERN_PRBS, 0x6000, 0x0000 } }, // 2^15-1   (taps [14,15])  // verified w/ T-Berd in T1 mode
  { "2^20-1",   { PATTERN_PRBS, 0x0000, 0x0009 } }, // 2^20-1   (taps [17,20])  // verified w/ T-Berd in T1 mode
  { "2^23-1",   { PATTERN_PRBS, 0x0000, 0x0042 } }, // 2^23-1   (taps [18,23])  // verified w/ T-Berd in T1 mode

  // This is a 2^20-1 pseudo random pattern that is modified with a 14 consecutive zero constraint
  { "QRSS",     { (14 << 8)|PATTERN_ZERO_SUPRESSED|
                  PATTERN_PRBS, 0x0000, 0x0009 } }, // QRSS     (taps [17,20])  // verified w/ T-Berd in T1 mode

  // Static patterns
  { "MARK (All Ones)",    { 0, 0xFFFF, 0x00FF } }, // MARK
  { "SPACE (All Zeros)",  { 0, 0x0000, 0x0000 } }, // SPACE

  //Random bits stream
//{ "Name",               {Type, Count of Ones bits, Bitlength of pattern (must be 2, 4, 8 or 24) } }
  { "1:1",                { PATTERN_RBITS, 1, 2 } }, // 1:1
  { "1:3",                { PATTERN_RBITS, 1, 4 } }, // 1:3
  { "3:1",                { PATTERN_RBITS, 3, 4 } }, // 3:1
  { "1:7",                { PATTERN_RBITS, 1, 8 } }, // 1:7
  { "7:1",                { PATTERN_RBITS, 7, 8 } }, // 7:1
  { "3-IN-24",            { PATTERN_RBITS, 3, 24 } }, // 3-IN-24
  { "2-IN-8",             { PATTERN_RBITS, 2, 8 } }, // 2-IN-8 
  //Static
  { "1:1 (Static)",                { 0, 0xAAAA, 0x00AA } }, // 1:1
  { "1:3 (Static)",                { 0, 0x8888, 0x0088 } }, // 1:3
  { "3:1 (Static)",                { 0, 0xEEEE, 0x00EE } }, // 3:1
  { "1:7 (Static)",                { 0, 0x8080, 0x0080 } }, // 1:7
  { "7:1 (Static)",                { 0, 0xFEFE, 0x00FE } }, // 7:1
  { "3-IN-24 (Static)",            { 0, 0x0004, 0x0044 } }, // 3-IN-24  (0x440004   0100 0100 0000 0000 0000 0100)
  { "2-IN-8 (Static)",             { 0, 0x4242, 0x0042 } }, // 3-IN-24  (0x424242   0100 0010 0100 0010 0100 0010)
  { "Autodetect",             { PATTERN_AUTODETECT, 0, 0 } }, // 3-IN-24  (0x424242   0100 0010 0100 0010 0100 0010)
};

BERTPATTERN demoBertPattern;
BERTPATTERN demoAutodetectedPattern;

/*******************************************************************************

  @func
    Initalize the global data for the Bert Sender and Receiver devices

  @comm

*******************************************************************************/
void DemoInitializeGlobalData(void)
{
  OTX_INT32 i, j, n, TxLi, RxLi, nTs;
  OTX_UINT32 nDivisor;

  OTX_UINT32 LiTable[8] = {
    0, 1, 2, 3, 4, 5, 6, 7    // To itself     

  };

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     g_coTimeSlotMask = 32;
     break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     g_coTimeSlotMask = 128;
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     OtxBertExReleaseLib();
     exit(1);
     break;
  }

  nDivisor = g_coTimeSlotMask / 32; // Number of E1 spans in one Burst highway

  if (g_demoEnv.coHighways * nDivisor == 1) {
    // One channel board. Loop it to itself
    LiTable[0] = 0;
  }

  n = 0;
  DemoLogf("Prepeare timeslots mapping data (used by burst device) :\n\n");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    TxLi = i % (g_demoEnv.coHighways * nDivisor); 
    RxLi = LiTable[TxLi] % (g_demoEnv.coHighways * nDivisor);

    for (j = 0; j < 1; j++,n++) {
      // Timeslot in E1 span is the same both for Tx and Rx
      nTs = 1 + j; 
      if (nTs > 31) {
        continue;
      }

      g_sqDspBertSender[n].nLi = TxLi;
      g_sqDspBertReceiver[n].nLi = RxLi;

      // The difference is its offset in particular burst highways 
      g_sqDspBertSender[n].nTs = nTs + (TxLi % nDivisor) * 32; // Add offset of span among burst device hw
      g_sqDspBertReceiver[n].nTs = nTs + (RxLi % nDivisor) * 32;

      g_sqDspBertSender[n].nBurst = TxLi / nDivisor;
      g_sqDspBertReceiver[n].nBurst = RxLi / nDivisor;
    }
  }

  DemoLogf("Timeslots mapping complete\n");

  demoBertPattern.nFlags = PATTERN_AUTODETECT;
  demoBertPattern.nPatHi = 0;
  demoBertPattern.nPatLo = 0;
}


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);
}

  
void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     DemoLogf("Opening Sleipnir board\n");
     break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     DemoLogf("Opening Thor-8-PCI Plus board\n");
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     exit(1);
     break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i * 4, g_hEventQueue, &g_hPhysDevs.hLi[j + i * 4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // Open the Line interfaces
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hThorController, result);
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hThorController, result);
}

  
/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
    // T1
    // Configure T1 Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    // Set them in T1 mode
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  } else {
    // E1
    // Configure E1 Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    // Set them in E1 mode
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }
}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

  @comm

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(void)
{
  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];


  // Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    OTX_INT32 nTs;

    // we must define the highway to use [0-1]
    iHighwayNo = nLi / (g_demoEnv.coLi / g_demoEnv.coHighways);

    // Create pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    DemoLogf("Creating pipe for Burst Sender/Receiver\n");

    for (nTs = 0; nTs < 32; nTs++) {
      sqTimeSlotMask[nTs] = 0xFF;

      // for this demo, we'll lay out the burst time slot mask as follows:
      //    [ 0..31]  - LI0:TS0-TS31 and LI4:TS0-TS31
      // OTX_DEVICE_THOR_8_PCI_PLUS only:
      //    [32..63]  - LI1:TS0-TS31 and LI5:TS0-TS31
      //    [64..95]  - LI2:TS0-TS31 and LI6:TS0-TS31
      //    [96..127] - LI3:TS0-TS31 and LI7:TS0-TS31
      //
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0-1 - one highway per each QuadFalc (128TS per highway)
      // as well as the size of the highway (128 TS)
      //
      // this is just a simple algorithm to accomplish the above
      sqBurstTimeSlotMask[nTs + ((nLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) * 32)] = 0xFF;
    }


    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe representing the external T1/E1
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64 * 32, &g_hPipes.hLiIn[nLi]);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvCreatePipe(64 * 32, &g_hPipes.hLiOut[nLi]);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    // -------------------------
    // Connect pipe: Burst -> Li
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_INPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, (g_demoEnv.coLi / g_demoEnv.coHighways) * 32);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi], -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    // -------------------------
    // Connect pipe: Li -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT, g_hPhysDevs.hLi[nLi], -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, (g_demoEnv.coLi / g_demoEnv.coHighways) * 32);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);
  }
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

  @comm

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS* pE1CfgData, 
  OUT OtxT1ConfigOptionsS* pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits = 0x3;
  pE1CfgData->bfSaBits = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe = OTX_FALSE;
  pE1CfgData->bAutoResync = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pE1CfgData->nLosSensitivity = 128;
  pE1CfgData->nLosRecovery = 128;
  pE1CfgData->nLineLength = 10;
  pE1CfgData->bHighTxPower = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion = OTX_FALSE;
  pT1CfgData->eLfaSensitivity = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pT1CfgData->nLosSensitivity = 128;
  pT1CfgData->nLosRecovery = 128;
  pT1CfgData->nLineLength = 10;
  pT1CfgData->bHighTxPower = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer = OTX_TRUE;
  pT1CfgData->bfClearChannels = 0x00;
}

void BertExCallback(OtxEventDataS *pEvent, void *pParam)
{
  DemoPrintBertEvent(pEvent);
}


/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i, n;


  for (i = 0; i < 1; i++) {
    result = OtxBertExCreate((OtxLogicalDeviceTypeE) OTX_LDEVICE_DATA_BERT_SENDER, OTX_INVALID_HANDLE_VALUE, i, &g_sqDspBertSender[i].hHandle);  DemoCheckRc(g_sqDspBertSender[i].hHandle, result);
    result = OtxBertExCreate((OtxLogicalDeviceTypeE) OTX_LDEVICE_DATA_BERT_RECEIVER, OTX_INVALID_HANDLE_VALUE, i, &g_sqDspBertReceiver[i].hHandle);  DemoCheckRc(g_sqDspBertReceiver[i].hHandle, result);
    result = OtxBertExSetCallbackFunc(g_sqDspBertReceiver[i].hHandle, BertExCallback,NULL); DemoCheckRc(g_sqDspBertReceiver[i].hHandle, result);
    //Fill burst device mapping table
    n = g_sqDspBertSender[i].nBurst;    g_hLogicDevs.mBurstSender[n].hDev[g_hLogicDevs.mBurstSender[n].nDevs++] = g_sqDspBertSender[i].hHandle;
    n = g_sqDspBertReceiver[i].nBurst;  g_hLogicDevs.mBurstReceiver[n].hDev[g_hLogicDevs.mBurstReceiver[n].nDevs++] = g_sqDspBertReceiver[i].hHandle;
  }
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectLogicalDevices(void)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[128];
  
    OTX_INT32 j;
    OTX_INT32 noOfTimeSlots = 31;

    for (i = 0; i < 1; i++) {
      // Timeslot on the E1 span
      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      for (j = 0; j < noOfTimeSlots; j++) {
        sqTimeSlotMask[g_sqDspBertSender[i].nTs + j] = 0xFF;
      }
      result = OtxBertExSetTimeSlotMask(g_sqDspBertSender[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      DemoCheckRc(g_sqDspBertSender[i].hHandle, result);

      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      for (j = 0; j < noOfTimeSlots; j++) {        
        sqTimeSlotMask[g_sqDspBertReceiver[i].nTs + j] = 0xFF;
      }
      result = OtxBertExSetTimeSlotMask(g_sqDspBertReceiver[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      DemoCheckRc(g_sqDspBertSender[i].hHandle, result);
    }
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

  @comm

*******************************************************************************/
void DemoInitializeLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 1; i++) {
    // Initialize the Bert Sender 
    result = OtxBertExDataSenderInit(
      g_sqDspBertSender[i].hHandle
    ); 
    DemoCheckRc(g_sqDspBertSender[i].hHandle, result);          

    // Initialize the Bert Receiver 
    result = OtxBertExDataReceiverInit(
      g_sqDspBertReceiver[i].hHandle
    ); 
    DemoCheckRc(g_sqDspBertReceiver[i].hHandle, result);              
  }
}
  

/*******************************************************************************

  @func
    Enable Configured Logical Devices

  @comm

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;  

  for (i = 0; i < 1; i++) {
    result = OtxBertExEnable(g_sqDspBertSender[i].hHandle); DemoCheckRc(g_sqDspBertSender[i].hHandle, result);    
  }

  for (i = 0; i < 1; i++) {
    result = OtxBertExEnable(g_sqDspBertReceiver[i].hHandle); DemoCheckRc(g_sqDspBertReceiver[i].hHandle, result);    
  }
}

/*******************************************************************************

  @func
    Configure BERT - Set pattern.

  @comm

*******************************************************************************/
void DemoSetBertPattern(BERTPATTERN* pattern)
{
  OTX_RESULT result;
  OTX_INT32 i;  

  for (i = 0; i < 1; i++) {    
    result = OtxBertExSetPattern(g_sqDspBertSender[i].hHandle, pattern);
  }

  for (i = 0; i < 1; i++) {    
    result = OtxBertExSetPattern(g_sqDspBertReceiver[i].hHandle, pattern);
  }
}

WORD Nibble2Text(BYTE b)
{
  return b < 10 ? '0' + b : ('A' + b) - 10;
}

/*******************************************************************************

  @func
    Print User-friendly pattern name

  @comm

*******************************************************************************/
void DemoPrintPatternInfo(BERTPATTERN* pattern) {  
  char patternName[128];
  if ((pattern->nFlags & PATTERN_PRBS) && (pattern->nFlags != 0xFFFF)) {
    //PRBS_PATTERN
    bool bFind = 0;
    int i = 0;
    while (!bFind && i < sizeof(arrPattern)/sizeof(arrPattern[0])) {
      if ((arrPattern[i].DspBertPattern.nFlags == pattern->nFlags)
          && (arrPattern[i].DspBertPattern.nPatHi == pattern->nPatHi)
          && (arrPattern[i].DspBertPattern.nPatLo == pattern->nPatLo))
        bFind = 1;
      else i++;
    }
    if (bFind) {
      sprintf(patternName, "PRBS: %s", arrPattern[i].sDescription);      
    } else {      
      sprintf(patternName, "PRBS, shift table: 0x%hc%hc%hc%hc%hc%hc", Nibble2Text(pattern->nPatHi >> 4),
                                                       Nibble2Text(pattern->nPatHi & 0x0F),
                                                       Nibble2Text((pattern->nPatLo & 0x00FF) >> 4),
                                                       Nibble2Text((pattern->nPatLo & 0x00FF) & 0x0F),
                                                       Nibble2Text(pattern->nPatLo >> 8 >> 4),
                                                       Nibble2Text(pattern->nPatLo >> 8 & 0x0F)
                                                       );
    }
  } else if (pattern->nFlags & PATTERN_RBITS) {
    //Random bitstream
    if ((pattern->nPatLo == 1) || (pattern->nPatHi-pattern->nPatLo == 1))
      sprintf(patternName, "Random Bit Stream: %i:%i", pattern->nPatLo, pattern->nPatHi-pattern->nPatLo);
    else sprintf(patternName, "Random Bit Stream: %i-IN-%i", pattern->nPatLo, pattern->nPatHi);
  } else {
    //Static
    sprintf(patternName, "Static: 0x%hc%hc%hc%hc%hc%hc", Nibble2Text(pattern->nPatHi >> 4),
                                                       Nibble2Text(pattern->nPatHi & 0x0F),
                                                       Nibble2Text((pattern->nPatLo & 0x00FF) >> 4),
                                                       Nibble2Text((pattern->nPatLo & 0x00FF) & 0x0F),
                                                       Nibble2Text(pattern->nPatLo >> 8 >> 4),
                                                       Nibble2Text(pattern->nPatLo >> 8 & 0x0F)
                                                       );
  }
  DemoLogf(patternName);  
}

/*******************************************************************************

  @func
    Prints a BERT Event.

  @comm

*******************************************************************************/
int DemoPrintBertEvent(OtxEventDataS* pEventData)
{
  DemoLogf("================== ");
  switch (pEventData->m_nCode) {
    case BERT_NOTIFY_SYNC: DemoLogf("BERT Event: Sync recieved.\n");
      break;
    case BERT_NOTIFY_UNSYNC: DemoLogf("BERT Event: Unsync recieved.\n");
      if (demoBertPattern.nFlags == PATTERN_AUTODETECT) {
        DemoSetBertPattern(&demoBertPattern);
      }
      break;
    case BERT_NOTIFY_ERROR: DemoLogf("BERT Event: Error recieved.\n");
      break;
    case BERT_NOTIFY_AUTODETECTED: 
      {        
        memcpy(&demoAutodetectedPattern, &pEventData->m_nParam, sizeof(BERTPATTERN));
        delete (void*) pEventData->m_nParam;
        DemoLogf("BERT Event: Autodetected pattern: ");
        DemoPrintPatternInfo(&demoAutodetectedPattern);
        DemoLogf("\n");
      }
      break;
  }
  return 0;
}

/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char* formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}


/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.

  @comm

*******************************************************************************/
static void DemoCheckRc(
  IN OTX_HANDLE hDevice, 
  IN OTX_RESULT result
)
{
  if (OTX_FAILURE(result)) {
    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32 cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32 eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32 nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int) nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

  @comm

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

  @comm

*******************************************************************************/
void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm* ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;
#elif defined(WIN32)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
#endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", ptCurrentDayAndTime->tm_mon + 1, ptCurrentDayAndTime->tm_mday, ptCurrentDayAndTime->tm_hour, ptCurrentDayAndTime->tm_min, ptCurrentDayAndTime->tm_sec, nTimeStampMs.sUint32.nLow);
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
}


/*******************************************************************************

  @func
    Idle Function that check for Events from the OTX Driver

  @comm

*******************************************************************************/
void* DemoIdleFunction(
  void* lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;
}


/*******************************************************************************

  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.

  @comm

*******************************************************************************/
void* DemoStatsThread(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;

  OTX_UINT32 stat_Starts;          // Number of actual dma transfers initiated
  OTX_UINT32 stat_Completes;       // Number of actual dma transfers completed
  OTX_UINT32 stat_Partial;         // Number of partial transfer due to framemask
  OTX_UINT32 stat_Full;            // Number of count of max dma size transfers
  OTX_UINT32 stat_Aborts;          // Number of actual dma transfers aborts
  OTX_UINT32 stat_Amount;          // Number of amount of data transfered
  OTX_UINT32 stat_Errors;          // Number of count of processing errors
  OTX_UINT32 stat_OUruns;          // Number of amount of over/under runs
  OTX_UINT32 stat_OUrunsUser;      // Number of amount of over/under runs caused by user mode
  OTX_UINT32 stat_DmaBusy;         // Number of count of DMA busy occurances
  OTX_UINT32 stat_Notify;          // Number of logical device notifies
  OTX_UINT32 stat_NotifyPartial;   // Number of logical device notifies - partial

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


void OnSendBurstData(struct sBurstSender* pSnd, OTX_UINT8* pBuf, OTX_UINT32 cbBuf)
{
  OTX_UINT32 nUsed = 0;                                                  
  OTX_UINT32 i;
  OTX_RESULT res;

  for (i = 0; i < cbBuf; i++)
    pBuf[i] = (i % 32) ? 0x7e : 0x7e;

  for (i = 0; i < pSnd->nDevs; i++) {
    nUsed = 0;
    res = OtxBertExProcessIO(pSnd->hDev[i], pBuf, cbBuf, &nUsed); // We have data in buffer
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
  }
}

void OnGetBurstData(struct sBurstReceiver* pRcv, OTX_UINT8* pBuf, OTX_UINT32 cbBuf)
{
  OTX_UINT32 nUsed = 0;                                                  
  OTX_UINT32 i;
  OTX_RESULT res;

  for (i = 0; i < (OTX_UINT32) pRcv->nDevs; i++) {
    nUsed = 0;
    res = OtxBertExProcessIO(pRcv->hDev[i], pBuf, cbBuf, &nUsed); // We have data in buffer
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
  }
}

/*******************************************************************************

  @func
    Burst Sender Thread Function

  @comm

*******************************************************************************/
void* DemoBurstSender(
  void* lpParameter   // thread data
)
{
  //clock_t     startTime, stopTime;
  OTX_RESULT otxRc;
  OTX_UINT32 nSourceId;
  OTX_UINT32 nTaskRef = 0;
  OTX_UINT32 nCount;
  OTX_UINT32 ui32;
  char* p;
  OTX_UINT32 coDataBlock = g_demoEnv.coLi / g_demoEnv.coHighways * 32 * DEMO_TX_SAMPLES;
  OTX_UINT8* pData;

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId < 0 || nSourceId> 2) {
    DemoLogf("DemoBurstSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    return 0;
  } {
    g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_TRUE;

    // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf).
    g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = coDataBlock;

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)
    coDataBlock = (coDataBlock / (g_demoEnv.coLi / g_demoEnv.coHighways)) * (g_demoEnv.coLi / g_demoEnv.coHighways);
    //   assert(coDataBlock % 128 == 0);
    pData = (OTX_UINT8 *) malloc(coDataBlock);

    memset(pData, 0x7e, coDataBlock);

    // until signaled to end, queue up and send the buffer
    while (g_hLogicDevs.mBurstSender[nSourceId].bActive) {
      p = (char *) pData;
      /*******************************************************/    
      OnSendBurstData(&g_hLogicDevs.mBurstSender[nSourceId], pData, coDataBlock);
      /*******************************************************/    
      for (nCount = coDataBlock; nCount > 0 && g_hLogicDevs.mBurstSender[nSourceId].bActive;) {
        ui32 = nCount;
        otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, p, &ui32);    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
        p += ui32;
        nTaskRef++;

        nCount -= ui32;
        if (nCount > 0) {
          OsaThreadSleep(10);
        }    // give the device a chance
        OsaThreadSleep(1);    // give the device a chance
      }
    }


    free(pData);
  }
  return 0;
}

/*******************************************************************************

  @func
    Burst Receiver Thread Function

  @comm

*******************************************************************************/
void* DemoBurstReceiver(void* lpParameter   // thread data
)
{
  clock_t startTime, stopTime;
  OTX_RESULT otxRc;
  OTX_UINT32 nSourceId;
  OTX_UINT8* buf;
  OTX_UINT32 nCount;
  OTX_UINT32 ui32;
  OTX_UINT32 coDataBlock = g_demoEnv.coLi / g_demoEnv.coHighways * 32 * DEMO_RX_SAMPLES;
  char* p;


  // validate parameters
  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId < 0 || nSourceId> 2) {
    DemoLogf("DemoBurstReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    return 0;
  }

  // our buffer size _MUST_ be a multiple of the maximum tineslots
  coDataBlock = ((coDataBlock / (g_demoEnv.coLi / g_demoEnv.coHighways)) * (g_demoEnv.coLi / g_demoEnv.coHighways));
  //  assert(coDataBlock % 512 == 0);
  buf = (OTX_UINT8 *) malloc(coDataBlock);
  startTime = clock();

  g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose = OTX_FALSE;
  g_hLogicDevs.mBurstReceiver[nSourceId].bActive = OTX_TRUE;

  // Pause receivers for senders to be ready and garbage data in DMA goes off
  OsaThreadSleep(1000); 

  while (g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    p = (char *) buf;
    for (nCount = coDataBlock; nCount > 0 && g_hLogicDevs.mBurstReceiver[nSourceId].bActive;) {
      ui32 = nCount;
      otxRc = OtxBurstReadData(g_hPhysDevs.hBurst, nSourceId, p, &ui32);   DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
      p += ui32;
      nCount -= ui32;
      if (nCount > 0) {
        OsaThreadSleep(10);    // give the device a chance
      }
      OsaThreadSleep(1);    // give the device a chance
    }
#if 1
    /*******************************************************/    
    OnGetBurstData(&g_hLogicDevs.mBurstReceiver[nSourceId], buf, coDataBlock);
    /*******************************************************/    
#endif
  }

  stopTime = clock();
  DemoLogf("DemoBurstReceiver[%d]: Done (time=%2.1f seconds).\n", nSourceId, (double) (stopTime - startTime) / CLOCKS_PER_SEC);

  free(buf);
  return 0;
}


/*******************************************************************************

  @func
    Prints a Driver Event.

  @comm

*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS* pEventData)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
   case OTX_DEVICE_LI_T1E1:
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);

     // Event from Li0 or Li1
     otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
     if (OTX_SUCCESS(otxRc)) {
       if (pEventData->m_nParam) {
         DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
       } else {
         DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
       }
     } else {
       DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
     }
     break;

   case OTX_DEVICE_BURST:
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);

     otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
     if (OTX_SUCCESS(otxRc)) {
       DemoLogf("Burst[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
     } else {
       DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
     }
     break;

   default:
     if (DemoPrintBertEvent(pEventData)) {
       break;
     }

     // Unhandled event
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);
     DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
     break;
  }
}


/*******************************************************************************

  @func

    Start of burst threads

  @comm

*******************************************************************************/
void DemoStartBurstThreads(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    g_hLogicDevs.mBurstReceiver[i].nSourceId = i;
    g_hLogicDevs.mBurstSender[i].nSourceId = i;

    OsaThreadStartEx(DemoBurstReceiver, (void *) &g_hLogicDevs.mBurstReceiver[i].nSourceId);
    OsaThreadStartEx(DemoBurstSender, (void *) &g_hLogicDevs.mBurstSender[i].nSourceId);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;
  static OTX_INT32 nCurrentHdlcDev = 0;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {
   case '?':
   case 'h':
   case 'H':
     DemoLogf("Cmd  Synopsis\n");
     DemoLogf("===  ========\n");
     DemoLogf(" ?   Display Help text\n");
     DemoLogf(" h   Ditto\n");
     DemoLogf(" q   Quit\n\n");
     DemoLogf("Related to T1/E1  ----------------------\n");
     DemoLogf(" 0-7 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
     DemoLogf(" l   Query Line Status of both Lis\n");
     DemoLogf(" p   Change current pattern\n\n");     
     break;   
      
   case ',':
      {
       OTX_INT32 nLi;

       for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
         DemoLogf("Reseting Li #%d... ", nLi);
         result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
         DemoLogf("Done.");
       }
     }
     break;


   case 'l':
      {
       OtxT1E1Layer1StatusE eLayer1Status;
       OTX_INT32 nLi;

       for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
         DemoLogf("Line status of Li #%d is ", nLi);
         result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
         DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
       }
     }
     break;

   case '0':
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
     /*
     Not support in this release

     if (commandChar - '0' < g_demoEnv.coLi) {
       // Select current Li
       g_demoEnv.nCurrentLi = commandChar - '0';
       DemoLogf("LI#%d Selected ....\n", g_demoEnv.nCurrentLi);
     }*/
     break;

   case '*':
      {
       OTX_UINT32 nValue;
       char sValue[80];

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
       OsaConsoleGetStr(sValue);
       if (sValue[0] == '\0') {
         break;
       }
       nValue = strtol(sValue, NULL, 0);

       // before setting the new trigger value, the burst device must not be enabled
       result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
       result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       DemoLogf("New queue notification value is %d\n", nValue);

       break;
     }

   case 'p':
     {
       int i;

       DemoLogf("Select new pattern:\n");
       for (i = 0; i < sizeof(arrPattern)/sizeof(arrPattern[0]); i++) {
         DemoLogf("%i. %s\n", i+1,arrPattern[i].sDescription);
       }
       DemoLogf("Your choice: ");
       i = 0;
       char in[3];
       while (i < 3 && in[i] != 13) {
         in[i] = OsaConsoleGetCh();      
         if ((in[i] >= '0') && (in[i] <= '9')) {
           if (in[i] != 13)
             DemoLogf("%c", in[i]);
           i++;
         }
       }
       DemoLogf("\n");
       {
         int newPattern = atoi(in);
         if (newPattern > 0) {
           newPattern--;
           if (newPattern < sizeof(arrPattern)/sizeof(arrPattern[0])) {
             demoBertPattern = arrPattern[newPattern].DspBertPattern;
             DemoLogf("New pattern: %s\n", arrPattern[newPattern].sDescription);
             DemoSetBertPattern(&demoBertPattern);
           } else {
             DemoLogf("Wrong selection.\n");
           }
         }
       }
     }
     break;

   case 'q':
   case 'Q':
     return 1;
     break;

   case '~':
      {
       char strBuf[12];
       OTX_INT32 nSleepTime;

       DemoPrintf("Enter milliseconds to sleep :");
       OsaConsoleGetStr(strBuf);

       if (strBuf[0]) {
         nSleepTime = strtol(strBuf, NULL, 0);
         DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
         OsaThreadSleep(nSleepTime);
       }
     }
     break;

   default:
     DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
     break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP)
{
  short i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr = NULL;
  envP->bExitOnError = OTX_FALSE;
  envP->eT1E1Mode = OTX_T1E1_MODE_E1;
  envP->boardNo = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi = 8;
  envP->coHighways = 2;
  envP->nCurrentLi = 0;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {
    // Scan command line arguments
    if (*(argv[i] + 1) == 'Z') {
      // Command input string -- case sensitive
      envP->pszInputStr = argv[i] + 2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {
        // Is argument a legal switch?
        if (*(argv[i] + 1) == 'L') {
          // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i] + 1) == 'Q') {
          // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'B') {
          // Board number
          envP->boardNo = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'T') {
          // Board Type
          if (*(argv[i] + 2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
            envP->coHighways = 2;
          } else if (*(argv[i] + 2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coHighways = 1;
//            envP->coHdlcPerLi = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'M') {
          // Li Mode
          if (*(argv[i] + 2) == 'E') {
            // E1
            envP->eT1E1Mode = OTX_T1E1_MODE_E1;
          } else if (*(argv[i] + 2) == 'T') {
            // T1
            envP->eT1E1Mode = OTX_T1E1_MODE_T1;
//            if (envP->nTimeSlots > 24) {
              //envP->nTimeSlots = 24;
            //}
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[P|S]      : Board Type used for the demo, where\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char* formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************

  @func

  prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf2(char* formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;
  for (i = 0; i < g_demoEnv.coHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

#if 0
  fclose(g_pBurstSendFile);
  fclose(g_pBurstRecvFile);
#endif

  fclose(g_pDemoLogFile);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char* argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("T8pDma, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for BERT device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

#if 0
  if ((g_pBurstSendFile = fopen("BurstSend.dat", "wt")) == NULL) {
    OsaConsoleExit();
    exit(-1);
  }

  if ((g_pBurstRecvFile = fopen("BurstRecv.dat", "wt")) == NULL) {
    OsaConsoleExit();
    exit(-1);
  }
#endif  
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
   {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static" : "dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();


  switch (OTX_RESULT_CODE(result)) {
   case OTX_RES_HANDLE:
     DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   case OTX_RES_NOT_SUPPORTED:
     DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   default:
     DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
     break;
  }

  result = OtxBertExConnectLib();

  switch (OTX_RESULT_CODE(result)) {
   case OTX_RES_HANDLE:
     DemoLogf("Error: Could not connect to the OTX BertEx library.\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   default:
     DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
     break;
  }

  // Init all lobal objects 
  DemoInitializeGlobalData();

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");


  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);
  DemoStartBurstThreads();

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n"); {
    OTX_INT32 i;


    g_bStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxBertExReleaseLib();
  OtxDrvReleaseLib();  

  return 0;
}
