/* $Workfile: OtxBertEx.h $   $Revision: 4 $    $Modtime: 11/09/04 3:16a $  */
/*******************************************************************************

  @doc PublicBertExLibApi OtxBertEx

  Copyright (C) 2004 Odin TeleSystems Inc., All Rights Reserved

  OTX Generic BertEx Library API

  @module
  OtxBert | Prototypes for the Bert Device (Sender/Receiver) Functions

  @contents2
  OtxBertX.h |

  @index func | OtxBertX

  @contents2
  OTX Logical Bert Sender/Receiver Functions - OtxBertExXXX() | The API functions for the Software High-level Data Link Controller (Bert) device

  @head3 OtxBertExXXX() Function Reference |

  @index func | OtxBertEx

*******************************************************************************/

#ifndef OTXBERTEX__LIB__H
#define OTXBERTEX__LIB__H

#include "OtxType.h"
#include "OtxErr.h"
#include "OtxDev.h"

#define OTX_LDEVICE_DATA_BERT_SENDER     -50
#define OTX_LDEVICE_DATA_BERT_RECEIVER   -51

// Structs for IOCtls
//Type of pattern
#define PATTERN_PRBS		        0x0001
#define PATTERN_INVERTED	      0x0002
#define PATTERN_RBITS           0x0010
#define PATTERN_AUTODETECT      0x0020

//Additinal flags for PATTERN_PRBS
#define PATTERN_ZERO_SUPRESSED	0x0004
#define PATTERN_ZERO_SUPR_MASK	0xff00    // # of consecutive zeros suppressed

//Event ID's
#define BERT_NOTIFY_SYNC     0x0091
#define BERT_NOTIFY_UNSYNC   0x0092
#define BERT_NOTIFY_ERROR    0x0093
#define BERT_NOTIFY_RX_LOST  0x0094
#define BERT_NOTIFY_TX_LOST  0x0095
#define BERT_NOTIFY_AUTODETECTED 0x0096

#if defined(__linux__)
  #define BERTX_API
#elif defined(WIN32)
  #ifdef OTX_BERTEXDLL_STATIC  
    #define OTXBERTEXDLL_API
  #else 
    #ifdef OTXBERTEXDLL_EXPORTS
      #define OTXBERTEXDLL_API __declspec(dllexport)
    #else
      #define OTXBERTEXDLL_API __declspec(dllimport)
    #endif
  #endif
#else
  #error "Platform is not supported."
#endif

  
#ifndef BERTX_API
  #define BERTX_API OTXBERTEXDLL_API
#endif

typedef OTX_UINT32 BERTX_HANDLE;

#define BERTX_INVALID_HANDLE_VALUE 0xFFFFFFFF

#pragma pack(push,8)

typedef void (*OtxExCallbackF)(OtxEventDataS *pEvent, void *pParam);

typedef struct _tagBertPattern
{
  unsigned short nFlags;
  unsigned short nPatLo;
  unsigned short nPatHi;
} BERTPATTERN, *PBERTPATTERN;

#pragma pack(pop)

#ifdef __cplusplus
extern "C" {
#endif

  // Library level access functions
  OTX_RESULT BERTX_API OtxBertExConnectLib();
  OTX_RESULT BERTX_API OtxBertExReleaseLib();

  /*******************************************************************************

    @func
       Creates software Bert-device 
       
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExCreate(IN  OtxLogicalDeviceTypeE eDevType, // @parm Class of device we are looking for, as known by the module in question. See also <t OtxLogicalDeviceTypeE>
    IN  OTX_HANDLE      hEventQueue, // @parm Handle to separate event queue (0 creates device specific event queue)
    IN  OTX_INT32       nEventSourceId, // @parm User specified ID to be used by this device as the Source ID for the events generated by the device
    OUT BERTX_HANDLE* phLogicDevice    // @parm Handle to the new logical device (if result code OTX_S_OK)
  );


  /*******************************************************************************

    @func
       Destroy Bert-device 
       
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExClose(BERTX_HANDLE hDevice);


  /*******************************************************************************

    @func
       Enable Bert-device 
       
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExEnable(BERTX_HANDLE hDevice);


  /*******************************************************************************

    @func
       Disable Bert-device 
       
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExDisable(BERTX_HANDLE hDevice);


  /*******************************************************************************

    @func
       Reset Bert-device 
       
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExReset(BERTX_HANDLE hDevice);


  OTX_RESULT BERTX_API OtxBertExSetTimeSlotMask(
    IN BERTX_HANDLE     hDevice, 
    IN OTX_UINT32       coTimeSlotMask, 
    IN OTX_UINT8* pnTimeSlotMask
  );


  /*******************************************************************************

    @func
       IO Process function

    @comm 
       Need to call it both for sender and receiver. Because of Soft devices 
       works with data buffers only, not with HW-devices.
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExProcessIO(IN BERTX_HANDLE hDevice, IN OTX_UINT8* pBertBuffer, IN OTX_UINT32   nBertBuffer, OUT OTX_UINT32* nBufUsed);


  /*******************************************************************************

    @func
      Initialize the Bert Data Sender.

    @comm
      Start Bert Data Sender operation with by calling the function <f OtxDrvEnable> and stop
      it by calling the function <f OtxDrvDisable>

    @xref
      <f OtxBertDataReceiverInit>
      <f OtxBertDataSendData>
      <e OtxLogicalDeviceTypeE.OTX_LDEVICE_DATA_Bert_SENDER>
      <t OtxDataFormatE>
 
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExDataSenderInit(
    IN BERTX_HANDLE     hBertDataDevice  // @parm Handle to logical Bert Data Device
  );


  /*******************************************************************************
  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExDataReceiverInit(
    IN BERTX_HANDLE     hBertDataDevice  // @parm Handle to logical Bert Data Device
  );


  /*******************************************************************************

    @func
      Set test pattern.

    @comm
      Usaully it is used in Bert Data Sender Thread


  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExSetPattern(
    IN BERTX_HANDLE     hBertDataDevice,   // @parm Handle to the Bert Data Sender Device
    IN BERTPATTERN     *pBertPattern
  );

  /*******************************************************************************

    @func
      Generate Error.

    @comm
      Usaully it is used in Bert Data Sender Thread


  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExInsertError(
    IN BERTX_HANDLE     hBertDataDevice,   // @parm Handle to the Bert Data Sender Device
    IN unsigned short   errorMask
  );


  /*******************************************************************************
    @mfunc

  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExSetCallbackFunc(
    IN BERTX_HANDLE     hBertDataDevice,   // @parm Handle to the Bert Data Sender Device
    IN OtxExCallbackF   pCallback,
    void               *pParam  
  );


  /*******************************************************************************
    @mfunc

  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExDelCallbackFunc(
    IN BERTX_HANDLE     hBertDataDevice,   // @parm Handle to the Bert Data Sender Device
    IN OtxExCallbackF   pCallback,
    void               *pParam  
  );

  /*******************************************************************************
    @mfunc

  *******************************************************************************/
  OTX_RESULT BERTX_API OtxBertExGetAutodetectedPattern(
    IN BERTX_HANDLE     hBertDataDevice,   // @parm Handle to the Bert Data Sender Device
    IN PBERTPATTERN     pPattern 
  );
#ifdef __cplusplus
} // extern "C"
#endif

/*--------------------------------------------------------------------------*\

  Copyright (C) Odin TeleSystems Inc., 2004


  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

\*--------------------------------------------------------------------------*/

#endif

